
def device_name(device_type, device_id):
    return '/{t}:{i}'.format(t=device_type, i=device_id)


def refined_tensor_or_graph_name(tensor_or_graph):
    import tensorflow as tf
    if isinstance(tensor_or_graph, tf.Tensor):
        return tensor_or_graph.name.replace(':', '_')
    return tensor_or_graph.name

def pre_work():
    from dxpy.learn.scalar import create_global_scalars
    create_global_scalars()


import tensorflow as tf
from ..utils.general import device_name
from .base import Model
from ..graph import Graph, NodeKeys


class MultiGPUSplitor(Model):
    """
    A Helper class to create
    """

    def __init__(self, name='gpu_splitor', nb_gpu=2):
        super().__init__(name, nb_gpu=nb_gpu)

    @classmethod
    def _default_config(cls):
        result = dict()
        result.update(super()._default_config())
        result.update({
            'lazy_create': True,
            'register_inputs': False,
            'register_outputs': False
        })
        return result

    def _kernel(self, feeds):
        x = feeds['input']
        return self._slice_tensor_for_gpu(x)

    def _get_shape_split(self, shape):
        if isinstance(shape, tf.TensorShape):
            shape_gpu = shape.as_list()
        else:
            shape_gpu = list(shape)
        if shape_gpu[0] is None:
            raise TypeError("Batch size can not be None for MultiGPUSplitor.")
        shape_gpu[0] = shape_gpu[0] // self.param('nb_gpu')
        return shape_gpu

    def _slice_tensor_for_gpu(self, tensor_input):
        result = []
        shape_gpu = self._get_shape_split(tensor_input.shape,)
        with tf.name_scope(self._get_tensor_name(tensor_input.name)):
            for id_slice in range(self.param('nb_gpu')):
                with tf.device(device_name('gpu', id_slice)):
                    slice_start = [id_slice * shape_gpu[0]] + \
                        [0] * (len(shape_gpu) - 1)
                    result.append(tf.slice(tensor_input, slice_start,
                                           shape_gpu, name='part_{}'.format(id_slice)))
        if len(result) == 0:
            return [tensor_input]
        return result

    def _get_tensor_name(self, name):
        prefix, idt = name.split(':')
        idt = int(idt)
        if idt == 0:
            return prefix
        else:
            return '{}_{}'.format(prefix, idt)


class PlaceHolder(Graph):
    """
    Placeholder for graph. Note this placeholder can be used to construct logic
    graph, thus may not create in tensorflow.
    """

    def __init__(self, shape, dtype=None, name='placeholder'):
        dtype = self._unified_dtype(dtype)
        if isinstance(shape, tf.TensorShape):
            shape = shape.as_list()
        super().__init__(name, shape=shape, dtype=dtype)

    @classmethod
    def _default_config(cls):
        from dxpy.collections.dicts import combine_dicts
        return combine_dicts({'dtype': tf.float32}, super()._default_config())

    @staticmethod
    def _unified_dtype(dtype):
        if isinstance(dtype, type(tf.float32)):
            return dtype
        type_desc = {
            tf.float32: ['FLOAT', 'FLOAT32', 'TF.FLOAT32'],
            tf.int64: ['INT', 'INT64', 'TF.INT64']
        }
        if isinstance(dtype, str):
            for k in type_desc:
                if dtype.upper() in type_desc[k]:
                    return k
            raise ValueError("Unknown dtype {}.".format(dtype))

    @property
    def shape(self):
        return self.param('shape')

    @property
    def dtype(self):
        return self.param('dtype')

    def as_tensor(self):
        if not NodeKeys.MAIN in self.nodes:
            self.register_main_node(tf.placeholder(self.dtype,
                                                   self.shape,
                                                   self.basename))
        return self.nodes[NodeKeys.MAIN]


class ShapeEnsurer(Model):
    """
    Used for assert on shapes or provide information for None dimensions.
    """

    def __init__(self, input_tensor, shape=None, *, batch_size=None, name='shape_ensurer'):
        super().__init__(name, inputs={NodeKeys.INPUT: input_tensor},
                         shape=shape, batch_size=batch_size, simple_output=True)

    def _kernel(self, feeds):
        with tf.name_scope('shape_ensurer'):
            if self.param('shape', raise_key_error=False) is None:
                shape = self.tensor(NodeKeys.INPUT).shape.as_list()
            else:
                shape = self.param('shape')
            return tf.reshape(feeds[NodeKeys.INPUT], self.param('shape'))



import tensorflow as tf

from .base import ScalarVariable


def batch_size():
    return graph().as_tensor()


def graph():
    if _instance is None:
        raise TypeError("Global batch size not created yet.")
    return _instance


_instance = None


class _BatchSize(ScalarVariable):
    def __init__(self):
        super().__init__('batch_size')


def create():
    global _instance
    _instance = _BatchSize('batch_size')




from functools import wraps

from flask import Flask, Response, make_response, request
from flask_restful import Api, Resource, reqparse


from ..exceptions import TaskNotFoundError
from ..import service


class TaskResource(Resource):
    def get(self, id):
        try:
            return Response(service.read_py(id), 200, mimetype="application/json")
        except TaskNotFoundError as e:
            return str(e), 404

    def put(self, id):
        try:
            task = request.form['task']
            return Response(service.update_py(task), 201, mimetype="application/json")
        except TaskNotFoundError as e:
            return str(e), 404

    def delete(self, id):
        try:
            return Response(service.delete_py(id), 200, mimetype="application/json")
        except TaskNotFoundError as e:
            return str(e), 404


class TasksResource(Resource):
    import json

    def get(self):
        task_jsons = []
        service.read_all_py().subscribe(lambda t: task_jsons.append(t))
        return Response(self.json.dumps(task_jsons), 200, mimetype="application/json")

    def post(self):
        task = request.form['task']
        res = service.create_py(task)
        return Response(self.json.dumps({'id': res}), 201, mimetype="application/json")


def add_api(api):
    from ..config import config as c
    from dxpy.web.urls import api_path
    api.add_resource(TaskResource, api_path(
        c['name'], '<int:id>', c['version'], c['base']))
    api.add_resource(TasksResource, api_path(
        c['names'], None, c['version'], c['base']))



import json
from functools import wraps

import requests
import rx

from ..exceptions import TaskDatabaseConnectionError, TaskNotFoundError


def connection_error_handle(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except requests.ConnectionError as e:
            raise TaskDatabaseConnectionError(
                "Task database server connection failed. Details:\n{e}".format(e=e))
    return wrapper


def req_url(tid=None):
    from ..config import config as c
    from dxpy.web.urls import req_url
    if tid is None:
        return req_url(c['names'], c['ip'], c['port'], None, c['version'], c['base'])
    else:
        return req_url(c['name'], c['ip'], c['port'], tid, c['version'], c['base'])


@connection_error_handle
def create(task_json):
    r = requests.post(req_url(), {'task': task_json})
    return r.json()['id']


@connection_error_handle
def read(tid):
    r = requests.get(req_url(tid))
    if r.status_code == 200:
        return r.text
    else:
        raise TaskNotFoundError(tid)


@connection_error_handle
def read_all():
    return rx.Observable.from_(json.loads(requests.get(req_url()).text))


@connection_error_handle
def update(task_json):
    task_json_dct = json.loads(task_json)
    r = requests.put(req_url(
        task_json_dct['id']), data={'task': task_json})
    if r.status_code == 404:
        raise TaskNotFoundError(task_json_dct['id'])


@connection_error_handle
def delete(tid):
    r = requests.delete(req_url(tid))
    if r.status_code == 404:
        raise TaskNotFoundError(tid)



# Author: Hong Xiang <hx.hongxiang@gmail.com>

"""
A time utility module, used for measure, record and analysis of time related tasks, all time in float are defined in
millisecond, this module defines the following classes:

- `TimeStamp`, a time stamp

Exception classes:

Functions:


"""

import json
import yaml
import datetime
import re
from dxpy.utils import Tags


def _parse_time_delta(s):
    if s is None:
        return None
    d = re.match(
        r'((?P<days>\d+) days?, )?(?P<hours>\d+):' +
        r'(?P<minutes>\d+):(?P<seconds>\d+)(\.(?P<microseconds>\d+))?',
        str(s)).groupdict(0)
    return datetime.timedelta(**dict(((key, int(value))
                                      for key, value in d.items())))


def now():
    return datetime.datetime.now()

class TimeStamps:
    """ A dict of time stamps
    """
    def __init__(self, time_stamps=None):
        if not isinstance(time_stamps, dict):
            raise TypeError("time_stamps must be a dict, not {}.".format(type(time_stamps)))
        for k in time_stamps:
            if not isinstance()
        self.time_stamps = time_stamps   


class TimeStamp(yaml.YAMLObject):
    """ A time stamp object.


    Members:

    - start: float, start time in secs,
    - run: float, run time in secs,
    - end: float, [Optional, Default=0.0], end time in secs

    Methods:


    """

    yaml_tag = '!time_stamp'

    DATETIMEFORMAT = "%Y-%m-%d %H:%M:%S"

    def __init__(self, start=None, run=None, end=None):
        self.start = start or now()
        self.run = run or now() - self.start
        self.end = end

    def now(self):
        return TimeStamp(self.start, end=self.end)

    @property
    def eta(self):
        if self.end > self.start:
            eta_now = self.end - self.start - self.run
            if eta_now < datetime.timedelta(0.0):
                return datetime.timedelta(0.0)
            else:
                return eta_now
        else:
            return None

    @classmethod
    def format_datetime(cls, dt):
        if dt is None:
            return Tags.undifined
        else:
            return datetime.datetime.strftime(dt, TimeStamp.DATETIMEFORMAT)

    @classmethod
    def format_duration(cls, td):
        return str(td)

    @classmethod
    def parse_datetime(cls, dt):
        if dt == Tags.undifined:
            return None
        return datetime.datetime.strptime(dt, TimeStamp.DATETIMEFORMAT)

    @classmethod
    def parse_duration(cls, td):
        return _parse_time_delta(td)

    @classmethod
    def to_yaml(cls, dumper, data):
        return yaml.MappingNode(cls.yaml_tag, value=[
            (yaml.ScalarNode(tag='tag:yaml.org,2002:str', value='start'),
             yaml.ScalarNode(tag='tag:yaml.org,2002:str', value=TimeStamp.format_datetime(data.start))),
            (yaml.ScalarNode(tag='tag:yaml.org,2002:str', value='run'),
             yaml.ScalarNode(tag='tag:yaml.org,2002:str', value=TimeStamp.format_duration(data.run))),
            (yaml.ScalarNode(tag='tag:yaml.org,2002:str', value='end'),
             yaml.ScalarNode(tag='tag:yaml.org,2002:str', value=TimeStamp.format_datetime(data.end)))
        ])

    @classmethod
    def from_yaml(cls, loader, node):
        data = loader.construct_mapping(node)
        start = TimeStamp.parse_datetime(data['start'])
        end = TimeStamp.parse_datetime(data['end'])
        run = TimeStamp.parse_duration(data['run'])
        return TimeStamp(start, run, end)


def estimate_end(ts, ratio):
    time_ela = ts.run
    time_total = time_ela / ratio
    time_end = ts.start + time_total
    return TimeStamp(ts.start, end=time_end)



import tensorflow as tf
from ..net import Net
from ..graph import NodeKeys

def sr_end(res, itp, ip_h, name='sr_end', is_res=True):
    """ Assuming shape(itp) == shape(ip_h)
    reps is center croped shape of itp/ip_h
    """
    with tf.name_scope(name):
        spo = res.shape.as_list()[1:3]
        spi = itp.shape.as_list()[1:3]
        cpx = (spi[0] - spo[0]) // 2
        cpy = (spi[1] - spo[1]) // 2
        crop_size = (cpx, cpy)
        itp_c = Cropping2D(crop_size)(itp)
        with tf.name_scope('output'):
            inf = add([res, itp_c])
        if is_res:
            with tf.name_scope('label_cropped'):
                ip_c = Cropping2D(crop_size)(ip_h)
            with tf.name_scope('res_out'):
                res_inf = sub(ip_c, inf)
            with tf.name_scope('res_itp'):
                res_itp = sub(ip_c, itp_c)
        else:
            res_inf = None
            res_itp = None
        return (inf, crop_size, res_inf, res_itp)


class SRMultiScale(Net):
    """
    Required inputs:
    'input/image2x', ... 'input/image8x'
    'label/image1x', ... 'label/image4x'
    """

    def __init__(self, inputs, name='network', **config):
        super().__init__(name, inputs, **config)

    @classmethod
    def _default_config(cls):
        from dxpy.collections.dicts import combine_dicts
        return combine_dicts({
            'nb_gpu': 2,
            'nb_down_sample': 3,
        }, super()._default_config())

    def _kernel(self, feeds):
        pass

    def _gpu_kernel(self, feeds):
        pass

    def _sr2x_kernel(self, input_images, label_images):
        



import click
from flask_restful import Api


@click.command()
def start():
    from flask import Flask
    from ..web import add_api
    from ...config import config as c
    """ start task database web service """
    app = Flask(__name__)
    api = Api(app)
    add_api(api)
    app.run(host=c['host'], port=c['port'], debug=c['debug'])



""" Base class definition """
import tensorflow as tf
from contextlib import contextmanager
from dxpy.filesystem import Path
from dxpy.collections.dicts import DXDict

# On restrict mode, tensors as allocated to CPU memory if not specified.
RESTRICT_MODE = True


class NodeKeys:
    EVALUATE = 'evaluate'
    INFERENCE = 'inference'
    TRAINER = 'trainer'
    SAVER = 'saver'
    LOSS = 'loss'
    MAIN = 'main'
    INPUT = 'input'
    LABEL = 'label'
    OUTPUT = 'output'
    MAIN_MODEL = 'main_model'
    CHILD_MODEL = 'child_model'
    TASK = 'task'


class Graph:
    """
    Base class of components.

    A `Graph` is an dict like collections of nodes and edges. Nodes of `Graph` can be a tensor another sub `Graph`.
    Edges representing relations of sub graphs/tensors, flow of information.

    A `Graph` is an generalization of `tf.Graph`, which is designed for following features:
        1. An unified interface of `tf.Graph` and general compute graph or operations/procedures;
        2. Seperate config and implementation, use TreeDict for configs, and supports multiple ways of config;
        3. An easy-to-use way of seperate/reuse subgraphs;
        4. Supports an warp of sessions.run/normal python function.
            Please add member method for tasks, and register them to tasks



    Methods:

    -   as_tensor(self):
        return self.nodes['main'], which is designed for sub_graphs.

    -   get_feed_dict(self, task=None):
        A method which returns a feed_dict, which can be used to update parent (in most cases, the graph which called 
        subgraph.get_feed_dict()) graph's get_feed_dict() or run task.
        Which is used to garantee output nodes (if is Tensor) to be valid under certain tasks, if task is None,
        a feed_dict should be provided so that all nodes are valid.

    -   run(self, task_name, feeds):
        Call a registered function. # TODO: make it different from directly call function, provide default feeds / unified feeds, etc.

    Properties:
        name: Path, name is used for:
            1. loading configs from global config object;
            2. its basename sometimes used for namescope/variable name in TensorFlow;
    """

    def __init__(self, name, **config):
        self.name = Path(name)
        self.c = self.__load_config(config)
        self.nodes = dict()
        self.tasks = dict()

    # Methods to be overrided:
    @classmethod
    def _default_config(cls):
        """ Override this method to provide default configs. """
        return dict()

    def _externel_feeds(self):
        return dict()

    def get_feed_dict(self, feeds=None, task=None):
        """
        Return feed dict for this graph to work for specific tasks.

        In most cases, it works as an translator for feeds to feed_dict,
        thus replacing name of feeds to tf.Tensor.
        """
        from dxpy.collections.dicts import combine_dicts
        result = self._externel_feeds()
        if feeds is None:
            return result
        for n in self.nodes:
            if n in feeds:
                result[self.tensor(n)] = feeds[n]
            if self.nodes[n] in feeds:
                result[self.tensor(n)] = feeds[self.nodes[n]]
        return result

    def _print_config_kernel(self, recursive, indent, fout):
        title = "{ind}>{cls}:{name}({fullname})".format(ind=" " * indent,
                                                        cls=__class__,
                                                        name=self.parse_name())
        indent_sub = indent + 4
        for k in self.nodes:
            if isinstance(self.nodes[k], tf.Tensor):
                print('{ind}tf.Tensor:{name}({sp})'.format(ind=" " * indent_sub,
                                                           name=k, sp=self.nodes[k].shape), file=fout)
            elif isinstance(self.nodes[k], Graph):
                if recursive:
                    self.nodes[k].print_config(recursive, fout, indent_sub)
                else:
                    print('{ind}Graph:{name}({sub_name})'.format(ind=" " * indent_sub,
                                                                 name=k, sub_name=self.nodes[k].name), file=fout)

    @property
    def basename(self):
        """
        Get the base name of graph name. Useful for variable_scope or name_scope of graph.
        """
        return self.name.name

    def register_node(self, name=None, tensor_or_subgraph=None):
        from .utils import refined_tensor_or_graph_name
        if tensor_or_subgraph is None:
            tensor_or_subgraph = name
            name = refined_tensor_or_graph_name(tensor_or_subgraph)
        self.nodes[name] = tensor_or_subgraph

    def register_task(self, name, func):
        self.tasks[name] = func

    def register_main_node(self, tensor_or_subgraph):
        self.register_node(NodeKeys.MAIN, tensor_or_subgraph)

    def register_main_task(self, func):
        self.register_task(NodeKeys.MAIN, func)

    def create_variable_node(self, dtype, shape, name, *, trainable=False, init_value=None):
        if init_value is not None:
            initer = tf.constant_initializer(init_value)
        else:
            initer = None
        self.register_node(name, tf.get_variable(
            name, shape, dtype, trainable=trainable, initializer=initer))
        return self.nodes[name]

    def create_placeholder_node(self, dtype, shape, name):
        self.register_node(name, tf.placeholder(dtype, shape, name))
        return self.nodes[name]

    def create_sub_graph_node(self, name, create_func):
        self.register_node(name, create_func())
        return self.nodes[name]

    def param(self, key, feeds=None, *,  default=None, raise_key_error=True):
        """
        Best practice: always use param instead of directly using self.c
        """
        if isinstance(feeds, dict) and key in feeds:
            return feeds[key]
        result = self.c.get(key, default)
        if result is None and raise_key_error:
            raise KeyError(key)
        return result

    def tensor(self, key=None):
        if key is None:
            return self.as_tensor()
        if not key in self.nodes:
            if isinstance(key, tf.Tensor):
                for n in self.nodes:
                    if self.nodes[n] == key:
                        return self.nodes[n]
            return None
        if isinstance(self.nodes[key], tf.Tensor):
            return self.nodes[key]
        elif isinstance(self.nodes[key], Graph):
            return self.nodes[key].as_tensor()
        return None

    def print_config(self, fout=None, recursive=False, indent=0):
        self._print_config_kernel(recursive, indent, fout)

    def __getitem__(self, name):
        if name in self.nodes:
            return self.nodes[name]
        elif name in self.tasks:
            return self.tasks[name]
        else:
            return None

    def __call__(self, feeds=None):
        return self.run(NodeKeys.MAIN, feeds)

    def run(self, task_name, feeds=None):
        return self.tasks[task_name](feeds)

    def as_tensor(self):
        return self.nodes.get(NodeKeys.MAIN)

    def __load_config(self, config_direct):
        from .config import config as config_global
        from dxpy.collections.dicts import get_hierarchy_dict, combine_dicts
        return combine_dicts(config_direct,
                             get_hierarchy_dict(config_global, self.name),
                             self._default_config())




import tensorflow as tf


def pet_image_super_resolution_dataset(dataset_name: str,
                                       image_type: str,
                                       batch_size: int,
                                       nb_down_sample: int,
                                       *,
                                       name: str='dataset',
                                       path_dataset: "str|None"=None):
    """
    Args:
        -   dataset_name: one of the following names:
            1.  analytical_phantoms
            2.  monte_carlo_phantoms [TODO: Impl]
            3.  mice [TODO: Imple]
            4.  mnist

        -   image_type: 'sinogram' or 'image'
        -   batch_size
    Returns:
        a `Graph` object, which has several nodes:
    Raises:
    """
    from ..petsino import PhantomSinograms
    from ...model.normalizer.normalizer import FixWhite, ReduceSum
    from ..super_resolution import SuperResolutionDataset
    from ...model.tensor import ShapeEnsurer
    from ...config import config
    config_origin = {

    }
    config_normalizer = {
        'mean': 4.88,
        'std': 4.36
    }
    config['dataset'] = {
        'origin': config_origin,
        'fix_white': config_normalizer
    }
    with tf.name_scope('{img_type}_dataset'.format(img_type=image_type)):
        dataset_origin = PhantomSinograms(name='dataset/origin',
                                          batch_size=batch_size)
        dataset_summed = ReduceSum('dataset/reduce_sum',
                                   dataset_origin['image'],
                                   fixed_summation_value=1e6).as_tensor()
        dataset = FixWhite(name='dataset/fix_white',
                           inputs=dataset_summed).as_tensor()
        dataset = SuperResolutionDataset('dataset/super_resolution',
                                         lambda: {'image': dataset},
                                         input_key='image',
                                         nb_down_sample=3)
    return dataset



""" 
A **Representation** of task
Since it is only a representation, it is not mutable, it has only properties.
No action is allowed.

Task fields:
    id,
    desc,
    workdir,
    worker,
    ttype,
    dependency,
    time_stamp,
    state,
    is_root,
    data
"""
import json
from enum import Enum
from dxpy.file_system.path import Path
from dxpy.time.timestamps import TaskStamp
from dxpy.time.utils import strf, strp, now
from dxpy.task.database import check_json


class State(Enum):
    BeforeSubmit = 0
    Pending = 1
    Runing = 2
    Complete = 3


class Worker(Enum):
    NoAction = 0,
    Local = 1,
    MultiThreading = 2,
    MultiProcessing = 3,
    Slurm = 4


class Type(Enum):
    Regular = 0,
    Command = 1,
    Script = 2


class Task:
    json_tag = '__task__'

    def __init__(self,
                 tid=None,
                 desc='',
                 workdir='.',
                 worker=None,
                 ttype=None,
                 state=None,
                 time_stamp=None,
                 dependency=None,
                 is_root=True,
                 data=None):
        self.id = tid
        self.desc = desc
        self.workdir = Path(workdir).abs
        if worker is None:
            worker = Worker.NoAction
        self.worker = worker
        if time_stamp is None:
            time_stamp = TaskStamp.create_now()
        self.time_stamp = time_stamp
        if dependency is None:
            dependency = []
        self.dependency = dependency
        if ttype is None:
            ttype = Type.Regular
        self.type = ttype
        if state is None:
            state = State.BeforeSubmit
        self.state = state
        self.is_root = is_root
        if data is None:
            data = {}
        self.data = data

    @property
    def is_pending(self):
        return self.state == State.Pending

    @property
    def is_before_submit(self):
        return self.state == State.BeforeSubmit

    @property
    def is_running(self):
        return self.state == State.Runing

    @property
    def is_complete(self):
        return self.state == State.Complete

    def command(self, generate_func=None) -> str:
        if generate_func is None:
            pass

    def to_json(self):
        return json.dumps(self.serialization(self))

    @classmethod
    def from_json(cls, s):
        check_json(s)
        return json.loads(s, object_hook=cls.deserialization)

    @classmethod
    def serialization(cls, obj):
        if isinstance(obj, Task):
            return {cls.json_tag: True,
                    'id': obj.id,
                    'desc': obj.desc,
                    'workdir': obj.workdir,
                    'worker': obj.worker.name,
                    'type': obj.type.name,
                    'state': obj.state.name,
                    'dependency': obj.dependency,
                    'time_stamp': {
                        'create': strf(obj.time_stamp.create),
                        'start': strf(obj.time_stamp.start),
                        'end': strf(obj.time_stamp.end)
                    },
                    'is_root': obj.is_root,
                    'data': obj.data, }
        raise TypeError(repr(obj) + " is not JSON serializable")

    @classmethod
    def deserialization(cls, dct):
        if cls.json_tag in dct:
            return Task(tid=dct['id'],
                        desc=dct['desc'],
                        workdir=dct['workdir'],
                        worker=Worker[dct['worker']],
                        ttype=Type[dct['type']],
                        state=State[dct['state']],
                        time_stamp=TaskStamp(
                            create=strp(dct['time_stamp']['create']),
                            start=strp(dct['time_stamp']['start']),
                            end=strp(dct['time_stamp']['end'])),
                        dependency=dct['dependency'],
                        is_root=dct['is_root'],
                        data=dct['data'])
        return dct

    def __str__(self):
        dct = self.serialization(self)
        dct['time_stamp'] = {'create': strf(self.time_stamp.create),
                             'start': strf(self.time_stamp.start),
                             'end': strf(self.time_stamp.end)}
        return json.dumps(dct, separators=(',', ':'), indent=4)

    def __hash__(self):
        return id(self)


def submit(task):
    task = Task.from_json(task.to_json())
    task.state = State.Pending
    return task


def start(task):
    task = Task.from_json(task.to_json())
    task.start = now()
    task.state = State.Runing
    return task


def complete(task):
    task = Task.from_json(task.to_json())
    task.end = now()
    task.state = State.Complete
    return task



from ..graph import Graph
from ..model.image import MultiDownSampler


class SuperResolutionDataset(Graph):
    """
    SuperResolutionDataset is a special kind of dataset which 
    """

    def __init__(self, name, origial_dataset_maker, input_key=None, nb_down_sample=None, with_shape_info=None, origin_shape=None, **config):
        super().__init__(name, input_key=input_key,
                         nb_down_sample=nb_down_sample,
                         with_shape_info=with_shape_info,
                         origin_shape=origin_shape,
                         **config)
        self._dataset_maker = origial_dataset_maker
        self.__construct()

    @classmethod
    def _default_config(cls):
        from dxpy.collections.dicts import combine_dicts
        return combine_dicts({'with_shape_info': False}, super()._default_config())

    def __down_sample_keys(self, start):
        return [('image{}x'.format(2**i), 2**i) for i in range(start, self.param('nb_down_sample') + 1)]

    def __construct(self):
        self.register_node('dataset',
                           self._dataset_maker())
        down_sample_ratios = {k: [v] * 2 for k,
                              v in self.__down_sample_keys(1)}
        origin_key = self.__down_sample_keys(0)[0][0]
        self.register_node('down_sampler',
                           MultiDownSampler(self.nodes['dataset'][self.param('input_key')],
                                            down_sample_ratios=down_sample_ratios,
                                            keep_original=True, original_key=origin_key,
                                            register_output_with_prefix=False))
        for k, _ in self.__down_sample_keys(0):
            self.register_node(k, self.nodes['down_sampler'][k])



def filter_out_none_elements(dct):
    result = dict()
    for n in dct:
        if dct[n] is not None:
            result[n] = dct[n]
    return result


def combine_dicts(*dcts, ignore_none_elements=True):
    result = dict()
    for e in reversed(dcts):
        if ignore_none_elements:
            e = filter_out_none_elements(e)
        result.update(e)
    return result


def get_hierarchy_dict(dct, path):
    from dxpy.filesystem import Path
    if isinstance(path, str):
        path = Path(path)
    result = dct
    for k in path.rel_parts():
        result = result.get(k)
        if result is None:
            return dict()
    return result



import unittest
from dxpy.task.model import creators, task


class TestCreators(unittest.TestCase):
    def test_create_command(self):
        t = creators.task_command(command='ls')
        self.assertEqual(t.data['command'], 'ls')
        self.assertEqual(t.type, task.Type.Command)

    def test_create_script(self):
        t = creators.task_script(file='run.sh', workdir='/tmp/test')
        self.assertEqual(t.data['file'], '/tmp/test/run.sh')
        self.assertEqual(t.type, task.Type.Script)

    def test_create_slurm(self):
        t = creators.task_slurm(file='run.sh', workdir='/tmp/test')
        self.assertEqual(t.data['file'], '/tmp/test/run.sh')
        self.assertEqual(t.type, task.Type.Script)
        self.assertEqual(t.worker, task.Worker.Slurm)



from flask import request
from flask_restful import Resource
from .model import Directory
from . import service
from ..exceptions import http_error


class DirectoryResource(Resource):
    def get(self, path):
        try:
            return Directory(path, 1).to_serializable(), 200
        except Exception as e:
            return http_error(e)

    def put(self, path):
        try:
            method = request.form.get('method')
            if method == 'cp':
                service.cp(Directory(path), request.form.get('target'))
            elif method == 'mv':
                service.mv(Directory(path), request.form.get('target'))
            return {'status': 'OK'}, 200
        except Exception as e:
            return http_error(e)

    def delete(self, path):
        try:
            service.rm(Directory(path))
            return {'status': 'OK'}, 200
        except Exception as e:
            return http_error(e)


class DirectoriesResource(Resource):
    def post(self):
        try:
            service.mkdir(Directory(request.form.get('path')))
            return {'status': 'OK'}, 201
        except Exception as e:
            return http_error(e)


def add_api(api, root, name='dir'):
    api.add_resource(DirectoryResource, root + '/{0}/<path>'.format(name))
    api.add_resource(DirectoriesResource, root + '/{0}s/<path>'.format(name))



import yaml
from functools import wraps
from .exceptions import UnknownConfigName

VERSION = 0.2

from .database import config as config_database

_default = {
    'name': 'task',
    'names': 'tasks',
    'version': 0.2,
}


class DatabaseConfigs:
    def __init__(self, file=None, root=None, ip=None, port=None, version=None, name=None, use_web_api=False):
        self.file = file or '/home/hongxwing/Workspace/databases/tasksv2.db'
        self.root = root or 'sqlite:///'
        self.ip = ip or '127.0.0.1'
        self.port = port or 23301
        self.echo = False
        self.version = version or VERSION
        self.name = name or 'taskdb'
        self.use_web_api = use_web_api
        self.debug = False

    @property
    def path(self):
        return self.root + self.file

    @property
    def task_url(self):
        return '/api/v{version}/{name}'.format(
            version=self.version, name=self.name)

    @property
    def tasks_url(self):
        return '/api/v{version}/{name}s'.format(
            version=self.version, name=self.name)


class InterfaceConfigs:
    def __init__(self, ip=None, port=None, name=None):
        self.ip = ip or '0.0.0.0'
        self.port = port or 23302
        self.name = name or 'task'
        self.version = VERSION
        self.debug = False

    @property
    def task_url(self):
        return '/api/v{version}/{name}'.format(
            version=self.version, name=self.name)

    @property
    def tasks_url(self):
        return '/api/v{version}/{name}s'.format(
            version=self.version, name=self.name)


CONFIGS = {
    'database': None,
    'interface': None,
}

CONFIGS_CLS = {
    'database': DatabaseConfigs,
    'interface': InterfaceConfigs,
}


def name_check(func):
    @wraps(func)
    def wrapper(name, *args, **kwargs):
        if name is not None and name not in CONFIGS:
            raise UnknownConfigName(name)
        return func(name, *args, **kwargs)
    return wrapper


@name_check
def get_config_cls(name):
    return CONFIGS_CLS[name]


@name_check
def get_config(name):
    if CONFIGS[name] is None:
        CONFIGS[name] = get_config_cls(name)()
    return CONFIGS[name]


@name_check
def set_config(name, yml_file=None, config=None):
    if yml_file is not None:
        CONFIGS[name] = yaml.load(yml_file)
    elif config is not None:
        CONFIGS[name] = config


@name_check
def set_config_by_name_key(name, key, value):
    c = get_config(name)
    setattr(c, key, value)


@name_check
def clear_config(name='all'):
    if name.upper() == 'ALL':
        for k in CONFIGS:
            CONFIGS[k] = None
    else:
        CONFIGS[name] = None



class NotDictError(TypeError):
    def __init__(self, true_type):
        super(__class__, self).__init__(
            "Need dict type, got {}.".format(true_type)
        )

class NotTreeDictError(TypeError):
    def __init__(self, true_type):
        super(__class__, self).__init__(
            "Need TreeDict type, got {}.".format(true_type)
        )


class KeyNotDictError(TypeError):
    def __init__(self, name, true_type):
        super(__class__, self).__init__(
            "Key {name} is supposed to be a dict, got {true_type}".format(name=name, true_type=true_type))



import tensorflow as tf

from ..graph import Graph


class ScalarVariable(Graph):
    def __init__(self, name, **config):
        super(__class__, self).__init__(name, **config)
        with tf.variable_scope(self.basename, reuse=self.c.get('reuse')):
            self.create_variable_node(
                tf.float32, [], 'value', init_value=self.c.get('init_value'))
            self.create_placeholder_node(tf.float32, [], 'new_value')
            self.assign_op = self.nodes['value'].assign(
                self.nodes['new_value'])
        self.register_node('setter', self.assign_op)
        self.register_main_node(self.nodes['value'])
        self.register_task('set', self.set_value)

    def set_value(self, feeds):
        tf.get_default_session().run(self.assign_op, feed_dict={
            self.nodes['new_value']: feeds})

    def get_value(self):
        return tf.get_default_session().run(self.as_tensor())



# import nose
# import tempfile
# import json
# import dxl.config as c

# def test_json_config_normal_call():
#         @o.json_config
#         def foo(a, b, *, c, d, e=4):
#             assert [a, b, c, d, e] == list(range(5))
#         foo(0, 1, c=2, d=3)

# def test_json_config_file():
#     with tempfile.NamedTemporaryFile(mode='w+') as fp:
#         fn = fp.name
#         a = {'b': 2, 'd': 4}
#         json.dump(a, fp)
#         fp.seek(0)
#         a_r = json.load(fp)        
#         @o.json_config
#         def foo(a, b, *, c, d, e=5):
#             assert [a, b, c, d, e] == list(range(1, 6))            
#         foo(1, c=3, config_files=fn)

# def test_json_config_redundent():
#     with tempfile.NamedTemporaryFile(mode='w+') as fp:
#         fn = fp.name
#         a = {'b': 2, 'd': 4, 'x': 100}
#         json.dump(a, fp)
#         fp.seek(0)
#         a_r = json.load(fp)        
#         @o.json_config
#         def foo(a, b, *, c, d, e=5):
#             assert [a, b, c, d, e] == list(range(1, 6))            
#         foo(1, c=3, config_files=fn)

# def test_auto_configs():
#     class A:
#         @o.auto_configs()
#         def __init__(self, a, b=None, c=2):
#             pass
#     a = A(4, 5, 6)
#     assert a.c.a == 4
#     assert a.c.b == 5
#     assert a.c.c == 6

# def test_auto_configs_defaults():
#     class A:
#         @o.auto_configs()
#         def __init__(self, a, b=None, c=2):
#             pass
#     a = A(4, 5)
#     assert a.c.a == 4
#     assert a.c.b == 5
#     assert a.c.c == 2

# def test_auto_configs_exclude():
#     class A:
#         @o.auto_configs(exclude=['b'])
#         def __init__(self, a, b=None, c=2):
#             pass
#     a = A(4, 5, 6)
#     assert a.c.a == 4
#     assert not a.c.has_key('b')
#     assert a.c.c == 6
    
# def test_auto_configs_include():
#     class A:
#         @o.auto_configs(include=['a'], exclude=['b'])
#         def __init__(self, a, b=None, c=2):
#             pass
#     a = A(4, 5, 6)
#     assert a.c.a == 4
#     assert not a.c.has_key('b')
#     assert not a.c.has_key('c')
    

# def test_auto_configs_inherit():
#     class A:
#         @o.auto_configs()
#         def __init__(self, a, b=None, *,  c=2):
#             pass
    
#     class B(A):
#         @o.auto_configs()
#         def __init__(self, d, *args, e, **kwargs):
#             super(B, self).__init__(*args, **kwargs)
#     b = B(3, a=0, b=1, c=2, e=4)
#     assert b.c.a == 0
#     assert b.c.b == 1
#     assert b.c.c == 2
#     assert b.c.d == 3
#     assert b.c.e == 4

# def test_combine_auto_json_config():
#     class A:
#         @o.json_config
#         @o.auto_configs()
#         def __init__(self, a, b=None, *, c, d=100):
#             pass
    
#     class B(A):
#         @o.json_config
#         @o.auto_configs()
#         def __init__(self, a, b, c, d, e):
#             super(B, self).__init__(a, b, c=c, d=d)

#     with tempfile.NamedTemporaryFile(mode='w+') as fp:
#         fn = fp.name
#         a = {'a': 100, 'b': 1, 'd': 3, 'c': 2, 'x': 100}        
#         json.dump(a, fp)
#         fp.seek(0)
#         a_r = json.load(fp)        
#         b = B(a=0, e=4, config_files=fn)
#         assert b.c.a == 0
#         assert b.c.b == 1
#         assert b.c.c == 2
#         assert b.c.d == 3
#         assert b.c.e == 4





from .database import TaskNotFoundError, TaskDatabaseConnectionError, InvalidJSONForTask


class ServiceNotFound(Exception):
    def __init__(self, name):
        super(__class__, self).__init__(
            "{name} service not found.".format(name=name))


class UnknownConfigName(Exception):
    def __init__(self, name):
        super(__class__, self).__init__(
            "Unknown config name: {name}.".format(name=name))


class UnknownTemplateNameError(Exception):
    def __init__(self, ukn_name):
        super(__class__, self).__init__("Unknown task template name {name}, suppoted names are {supp}.".format(
            name=ukn_name, supp=[k for k in TEMPLATE_NAME_MAP.keys()]))



import tensorflow as tf
import numpy as np
import random
from ..base import DatasetClassic
from ..preprocessing.normalizer import SelfMinMax
import warnings

warnings.warn(DeprecationWarning(
    "CLASSIC MNIST SHOULD NOT BE USED IN PRODUCTION SYSTEM."))


class MNISTClassic(DatasetClassic):
    def __init__(self, name='/dataset', **config):
        super(__class__, self).__init__(name, **config)
        self.images, self.labels = self.__load_data()
        if self.param('normalization')['method'].lower() == 'selfminmax':
            self._normalizer = SelfMinMax(self.name / 'normalization')

    @classmethod
    def _default_config(cls):
        return {
            'fields': {
                'image': {
                    'dtype': tf.float32,
                    'shape': [28, 28, 1],
                },
                'label': {
                    'dtype': tf.int64,
                    'shape': [10]
                }
            },
            'batch_size': 32,
            'normalization': {
                'method': 'selfminmax'
            },
        }

    def _load_sample(self, feeds=None):
        idx = random.randint(0, self.images.shape[0] - 1)
        image = np.reshape(self.images[idx, ...], [28, 28, 1])
        image = self._normalizer(image)['data']

        label = self.labels[idx, ...]
        return {'image': image, 'label': label}

    def __load_data(self):
        from tensorflow.examples.tutorials.mnist import input_data
        from tensorflow.contrib.data import Dataset
        mnist = input_data.read_data_sets(
            '/home/hongxwing/Datas/mnist/tfdefault', one_hot=True)
        images = mnist.train.images
        labels = mnist.train.labels

        return images, labels



import tensorflow as tf
from ..base import Model, NodeKeys, ModelPipe
from .. import activation


class Conv2D(Model):
    def __init__(self, name='conv2d', input_tensor=None, filters=None, kernel_size=None,
                 *, strides=None, padding=None, activation=None, **config):
        super().__init__(name, inputs={NodeKeys.INPUT: input_tensor}, kernel_size=kernel_size, strides=strides, padding=padding, filters=filters,
                         activation=activation, **config)

    @classmethod
    def _default_config(cls):
        from dxpy.collections.dicts import combine_dicts
        return combine_dicts({
            'kernel_size': 3,
            'strides': (1, 1),
            'padding': 'same',
            'reuse': False,
            'activation': 'basic',
            'pre_activation': False,
            'post_activation': False
        }, super()._default_config())

    def _kernel(self, feeds):
        x = feeds[NodeKeys.INPUT]
        acc = activation.unified_config(self.param('activation', feeds))
        activation.apply(acc, x, 'pre')        
        x = tf.layers.conv2d(x, self.param('filters', feeds),
                             self.param('kernel_size', feeds),
                             self.param('strides', feeds),
                             self.param('padding', feeds), name='convolution')
        activation.apply(acc, x, 'post')
        return x


class InceptionBlock(Model):
    def __init__(self, name='incept', input_tensor=None, **config):
        super().__init__(name, inputs={NodeKeys.INPUT: input_tensor}, **config)

    @classmethod
    def _default_config(cls):
        result = dict()
        result.update(super()._default_config())
        result.update({
            'paths': 3,
            'reuse': False,
            'pre_activation': True,
            'activation': 'relu'
        })
        return result

    def _kernel(self, feeds):
        x = feeds[NodeKeys.INPUT]
        filters = x.shape.as_list()[-1]
        if self.param('pre_activation', feeds):
            with tf.name_scope('pre_activation'):
                x = get_activation(self.param('activation', feeds))(x)
        paths = []
        for i_path in range(self.param('paths')):
            with tf.variable_scope('path_{}'.format(i_path)):
                h = Conv2D('conv2d_0', x,
                           filters=filters, kernel_size=1, pre_activation=True).as_tensor()
                for j in range(i_path):
                    h = Conv2D('conv2d_{}'.format(j + 1), h,
                               filters=filters, pre_activation=True).as_tensor()
                paths.append(h)
        with tf.name_scope('concat'):
            x = tf.concat(paths, axis=-1)
        x = Conv2D('conv_end', x, filters=filters).as_tensor()
        return x


class Residual(Model):
    def __init__(self, name, input_tensor, child_model=None, **config):
        if child_model is not None:
            child_models = {NodeKeys.CHILD_MODEL: child_model}
        else:
            child_models = None
        super().__init__(name,
                         inputs={NodeKeys.INPUT: input_tensor},
                         child_models=child_models,
                         **config)

    @classmethod
    def _default_config(cls):
        result = dict()
        result.update(super()._default_config())
        result.update({'ratio': 0.3})
        return result

    def _kernel(self, feeds):
        x = feeds[NodeKeys.INPUT]
        with tf.variable_scope('main_path'):
            y = self._child_models[NodeKeys.CHILD_MODEL](x)
        with tf.name_scope('add'):
            res = x + y * self.param('ratio', feeds)
        return res


class StackedConv2D(Model):
    def __init__(self, name, input_tensor, nb_stacked=None, activation=None, pre_activation=None, post_activation=None, **config):
        super().__init__(name, {NodeKeys.INPUT: input_tensor}, nb_stacked=nb_stacked,
                         activation=activation, pre_activation=pre_activation, post_activation=post_activation, **config)

    @classmethod
    def _default_config(cls):
        from dxpy.collections.dicts import combine_dicts
        cfg = {
            'activation': 'relu',
            'pre_activation': False,
            'post_activation': False,
            'filters': 32,
        }
        return combine_dicts(cfg, super()._default_config())

    def _pre_kernel_post_inputs(self):
        for i in range(self.param('nb_stacked')):
            if i == 0:
                pre_activation = self.param('pre_activation')
            else:
                pre_activation = False
            if i == self.param('nb_stacked') - 1:
                post_activation = self.param('post_activation')
            else:
                post_activation = True
            model = Conv2D('conv2d_{}'.format(i),
                           filters=self.param('filters'),
                           input_tensor=None,
                           activation=self.param('activation'),
                           pre_activation=pre_activation,
                           post_activation=post_activation,
                           lazy_create=True)
            self.register_child_model('conv2d_{}'.format(i), model)

    def _kernel(self, feeds):
        x = feeds[NodeKeys.INPUT]
        for i in range(self.param('nb_stacked')):
            x = self._child_models['conv2d_{}'.format(i)](x)
        return x



import tensorflow as tf
from ..model import Model
from ..graph import Graph, NodeKeys


class Net(Model):
    """ Base class of nets.
    Net add some special tasks based on graph:
        1. train
        2. inference
        3. evaluate
        4. save/load
    """

    def __init__(self, name, inputs=None, **config):
        super().__init__(name, inputs, **config)

    @classmethod
    def _default_config(cls):
        from dxpy.collections.dicts import combine_dicts
        return combine_dicts({
            'add_trainer': True,
            'add_saver': True
        }, super()._default_config())

    def _tensors_need_summary(self):
        return dict()

    def _post_kernel_post_outputs(self):
        super()._post_kernel_post_outputs()
        from ..train import Trainer, Saver
        if self.param('add_trainer'):
            if NodeKeys.LOSS in self.nodes:
                self.register_node(NodeKeys.TRAINER,
                                   Trainer(self.name / 'trainer', self.nodes[NodeKeys.LOSS]))
        if self.param('add_saver'):
            self.register_node(NodeKeys.SAVER, Saver(self.name / 'saver'))

    def post_session_created(self):
        pass

    @property
    def session(self):
        return tf.get_default_session()

    def train(self, feeds=None):
        return self.nodes[NodeKeys.TRAINER](feeds)

    def inference(self, feeds=None):
        return self.session.run(self.tensor(NodeKeys.INFERENCE), self.get_feed_dict(feeds))

    def evaluate(self, feeds=None):
        return self.session.run(self.tensor(NodeKeys.EVALUATE), self.get_feed_dict(feeds))

    def save(self, feeds=None):
        return self.nodes[NodeKeys.SAVER].run('save', feeds)

    def load(self, feeds=None):
        return self.nodes[NodeKeys.SAVER].run('load', feeds)



import click


@click.command()
@click.option('--name', '-n', type=str, help='Name of script to run.')
def shell(name):
    import subprocess
    from ..service import get_script
    if name is None:
        raise TypeError("Name of shell script can not be None.")
    path_file = get_script(name)
    subprocess.run(path_file.abs)



"""
Add yaml support for generic classes.
Currently support:
- subcless of Enum
"""
from ruamel.yaml import YAML
yaml = YAML()


def register(cls):
    yaml.register_class(cls)
# from enum import Enum


# def add_yaml_support_enum(cls, tag):
#     fmt = 'YAML support added for {cls} with tag: {tag}.'
#     logger.info(fmt.format(cls=cls, tag=tag))

#     def representer(dumper, data):
#         return dumper.represent_scalar(tag, data.name)
#     yaml.add_representer(cls, representer)

#     def constructor(loader, node):
#         value = loader.construct_scalar(node)
#         for v in cls:
#             if v.name == value:
#                 return v
#         raise ValueError(
#             "Unkown value {value} of type {cls}.".format(value=value, cls=cls))
#     yaml.add_constructor(tag, constructor)


# def add_yaml_support(cls, tag):
#     if issubclass(cls, Enum):
#         add_yaml_support_enum(cls, tag)



import os
from dxpy.time.utils import now


class DeamonService:
    @classmethod
    def cycle(cls):
        from .service import auto_complete, auto_submit_chain, auto_submit_root, auto_start
        auto_complete()
        auto_submit_chain()
        auto_submit_root()
        auto_start()
        print('Cycle at {t}.'.format(t=now(True)))

    @classmethod
    def start(cls, cycle_intervel=None):
        from apscheduler.schedulers.blocking import BlockingScheduler
        scheduler = BlockingScheduler()
        scheduler.add_job(cls.cycle, 'interval', seconds=60)
        print(
            'Press Ctrl+{0} to exit'.format('Break' if os.name == 'nt' else 'C'))
        try:
            cls.cycle()
            scheduler.start()
        except (KeyboardInterrupt, SystemExit):
            pass

    @staticmethod
    def stop():
        raise NotImplementedError

    @staticmethod
    def restart():
        raise NotImplementedError

    @staticmethod
    def is_running():
        pass



import tensorflow as tf

from ..graph import Graph


def keep_prob():
    return graph().as_tensor()


def graph():
    if _instance is None:
        raise TypeError("Global keep prob not created yet.")
    return _instance


def get_value():
    return tf.get_default_session().run(keep_prob())


def set_value(value):
    graph().run('set', value)


_instance = None


class _KeepProb(Graph):
    def __init__(self, name):
        super(__class__, self).__init__(name)
        with tf.variable_scope('keep_prob'):
            self.create_variable_node(tf.float32, [], 'value', init_value=1.0)
            self.create_placeholder_node(tf.float32, [], 'new_value')
            self.assign_op = self.nodes['value'].assign(
                self.nodes['new_value'])
        self.register_node('setter', self.assign_op)
        self.register_main_node(self.nodes['value'])
        self.register_task('set', self.set_value)

    def set_value(self, feeds):
        tf.get_default_session().run(self.assign_op, feed_dict={
            self.nodes['new_value']: feeds})


def create():
    global _instance
    _instance = _KeepProb('keep_prob')



import click


@click.group()
def files():
    pass


@files.command()
def sum():
    pass


@files.command()
def cat():
    pass


@files.command()
def hadd():
    pass


@files.command()
@click.argument('source', nargs=1, type=str)
@click.argument('target', nargs=1, type=str)
@click.argument('filenames', nargs=-1, type=str)
@click.option('-d', '--depth', type=int, default=0)
@click.option('-v', '--verbose', is_flag=True)
def fetch(source, target, filenames, depth, verbose):
    from ...api import files_in_directories
    from fs.osfs import OSFS
    from fs.copy import copy_file
    from fs.path import dirname
    with OSFS(source) as sor:
        with OSFS(target) as tar:
            files = files_in_directories(sor, ['*'], filenames, depth)
            for f in files:
                if not tar.exists(dirname(f)):
                    tar.makedirs(dirname(f))
                copy_file(sor, f, tar, f)
                if verbose:
                    click.echo('[COPY] {} => {}.'.format(sor.getsyspath(f),
                                                         tar.getsyspath(f)))


@files.group()
def clears():
    """
    Quick cleaner of prefab utiles
    """


# @clears.option("--name", "-n", type=str, help="Name of prefab utiles.")
# def pyc():
#     pass


@click.group()
def dirs():
    pass


@dirs.command()
def merge():
    pass


@dirs.command()
def sbatch():
    pass



from dxpy.filesystem import Path, File


def get_scripts_path():
    import dxpy
    dxpy_init_file = dxpy.__file__
    path_dxpy = Path(dxpy_init_file).father
    path_scripts = Path(path_dxpy) / 'scripts'
    return path_scripts


def get_script(name):
    path_scripts = get_scripts_path()
    possible_suffix = ['', '.sh', '.py']
    for s in possible_suffix:
        path_file = path_scripts / name + s
        if File(path_file).exists:
            break
    return path_file



from ..base import Model, ModelPipe, NodeKeys


class SuperResolutionKernel(ModelPipe):
    def __init__(self, reps, name='kernel', **configs):
        super().__init__(reps, name, **config)

    def _kernel(self, feeds):
        x = feeds[NodeKeys.INPUT]
        


class SRB(Model):
    def __init__(self, image_input, represents=None, name='srb', **configs):
        super().__init__(name=name, inputs={'ipt': image_input, 'rep': represents},
                         **configs)

    def _kernel(self, feeds):
        if feeds['rep'] is None:



"""
==================
Task system design
==================

Serivices
=========
TaskService:
    High level API.
    Dependency: 
        - StoreService (TaskJSON)        
    Output:
        TaskJSON

TaskServiceWeb:
    A RESTful API warper on TaskService.
    Dependency:
        - TaskService (TaskJSON)
    Output:
        TaskJSON

StoreService:
    Dependency:
        - DatabaseWebService (TaskJSON)
    Output:
        TaskPy, TaskJSON

RunService:
    Dependency:
        - TaskStoreService (TaskPy)
    Output:
        TaskPy
        
FactoryService:
    Dependency:
        - TaskStoreService (TaskPy)
    Output:
        TaskPy


Layers:



1:  RunService:
    Internal Representations:
        TaskPy, Task
0:  Database
    Interface representations:
        TaskJSON
    Internal Representations:
        TaskDB, TaskJSON
    Features:
        Managing Serilization of Tasks

    






Task Representations
====================
Overview:        
    All representations have full information.
Representations:
    TaskPy
        Python Object. Properties are objects ready to use
    TaskYAML
        Supports easy dump/load to TaskPython.
        Main serilization format of Tasks.
    TaskJSON
        JSON i/o. For simple communication only.
        Details are stored in filed 'body' as YAML.
    TaskDB
        Not really a representation, true resource of tasks. The only mutable representation.
Transform:
    TaskPy <=> TaskYAML
    TaskPy => TaskJSON    
    TaskJSON => TaskYAML
    TaskJSON <=> TaskDB


Overview
--------
There are four kinds of task objects in dxpy.task.
#. TaskDBModel
    Task resource
Task object is a *representation* of a logical task, which is designed to be:
#. 
#. 

Features
--------
#. self content
    A task object contains **ALL** information to perform a task with the help of task server.
#. subtask list
    A task may be consisted by multiple subtasks. There are two type of tasks, 
#. task state inspect
#. task serilization, database support (Optional)
    retrieve task from file, or database


Task server
-----------
#. Task factory
#. Task query
    #. on running
    #. history (recent)

#. List of tasks
#. Create(Load Tasks):
    #. create from scrach
    #. from file
        #. given path
        #. from database 




SLURM => Resource Unlimited
(Pending system)
OS, Hardware => Resource Limited
"""
from .api import *




from .. import service

class Database:
    from ..model import Database as DBM
    from ..config import config as c

    @classmethod
    def create(cls):
        cls.DBM.create()

    @classmethod
    def clear(cls):
        cls.DBM.clear()

    @classmethod
    def remove():
        from dxpy.filesystem import File, rm
        rm(File(c['path']))

    @classmethod
    def exists():
        from dxpy.filesystem import File
        return File(c['path']).exists


def create(s):
    from ..config import config as c
    if c['use_web_api']:
        return service.create_web(s)
    else:
        return service.create_py(s)


def read(tid):
    from ..config import config as c
    if c['use_web_api']:
        return service.read_web(tid)
    else:
        return service.read_py(tid)


def read_all():
    from ..config import config as c
    if c['use_web_api']:
        return service.read_all_web()
    else:
        return service.read_all_py()


def update(s):
    from ..config import config as c
    if c['use_web_api']:
        return service.update_web(s)
    else:
        return service.update_py(s)


def delete(tid):
    from ..config import config as c
    if c['use_web_api']:
        return service.delete_web(tid)
    else:
        return service.delete_py(tid)



import numpy as np


def unbatch(tensor):
    if isinstance(tensor, np.ndarray):
        return list(map(lambda x: x[0, ...], np.split(tensor, tensor.shape[0])))
    raise TypeError("numpy.ndarray is required, got {}.".format(type(tensor)))



import tensorflow as tf

from ..graph import Graph


def global_step():
    return graph().as_tensor()


def graph():
    if _instance is None:
        raise TypeError("Global step not created yet.")
    return _instance


def get_value():
    return tf.get_default_session().run(global_step())


def set_value(value):
    graph().run('set', value)


_instance = None


class _GlobalStep(Graph):
    def __init__(self, name='global_step'):
        super(__class__, self).__init__(name)
        self.register_main_node(tf.Variable(0, dtype=tf.int64, trainable=False,
                                            name='global_step'))
        with tf.name_scope('global_step_setter'):
            self.create_placeholder_node(tf.int64, [], 'new_value')
            self.assign_op = self.as_tensor().assign(self.nodes['new_value'])
        self.register_node('setter', self.assign_op)
        self.register_task('set', self.set_value)

    def set_value(self, feeds):
        tf.get_default_session().run(self.assign_op, feed_dict={
            self.nodes['new_value']: feeds})


def create():
    global _instance
    _instance = _GlobalStep()



from dxpy.filesystem.path import Path
from dxpy.configs.base import Configs
from ruamel.yaml import YAML


def save():
    return {'frequency': 100, 'method': 'step'}


def load():
    return {'is_load': True, 'step': -1}


def model_filesystem():
    return {'path_model': './model', 'ckpt_name': 'save'}


def train():
    return {'model_fs': model_filesystem(),
            'load': load(),
            'save': save()}



import os
from dxpy.filesystem import Path


def _default_dataset_path():
    return str(Path(os.environ.get('HOME')) / 'Datas')


config = {
    'dataset_root_path': os.environ.get('PATHS_DATASET', _default_dataset_path())}



import tensorflow as tf
from .base import Model, NodeKeys


def resize(inputs, size, method=None, name='upsampling'):
    if method is None:
        method = 'bilinear'
    input_shape = inputs.shape.as_list()
    h0, w0 = input_shape[1:3]
    h1 = int(h0 * size[0])
    w1 = int(w0 * size[1])
    with tf.name_scope(name):
        if method == 'nearest':
            h = tf.image.resize_nearest_neighbor(inputs, size=[h1, w1])
        elif method == 'bilinear':
            h = tf.image.resize_bilinear(inputs, size=[h1, w1])
    return h


def add_poisson_noise(input_tensor):
    if not isinstance(input_tensor, tf.Tensor):
        raise TypeError("Requre Tensor, got {}.".format(type(input_tensor)))
    with tf.name_scope('add_poisson_noise'):
        if input_tensor.dtype != tf.float32:
            input_tensor = tf.cast(input_tensor, tf.float32)
        return tf.random_poisson(input_tensor, [])


class DownSampler(Model):
    def __init__(self,
                 input_tensor, ratio,
                 *,
                 method=None, padding=None, name='down_sampler',
                 **config):
        super().__init__(name, inputs={NodeKeys.INPUT: input_tensor},
                         ratio=ratio, method=method, padding=padding, **config)

    @classmethod
    def _default_config(cls):
        from dxpy.collections.dicts import combine_dicts
        cfg = {
            'method': 'mean',
            'padding': 'same'
        }
        return combine_dicts(cfg, super()._default_config())

    def _kernel(self, feeds):
        x = feeds[NodeKeys.INPUT]
        if self.param('method') == 'mean':
            with tf.name_scope('down_sample'):
                x = tf.nn.avg_pool(x,
                                   [1] + list(self.param('ratio')) + [1],
                                   padding=self.param('padding').upper(),
                                   strides=[1] + list(self.param('ratio')) + [1])
        else:
            raise ValueError("Unsupported down sample method")
        return x


class MultiDownSampler(Model):
    """
    A helper model to create multiple down sampler and form their results to a dict.
    """

    def __init__(self, input_tensor, down_sample_ratios=None,
                 *,
                 keep_original=None,
                 original_key=None,
                 with_shape_info=None,
                 original_shape=None,
                 name='multi_down_sampler',
                 **config):
        super().__init__(name,
                         inputs={NodeKeys.INPUT: input_tensor},
                         down_sample_ratios=down_sample_ratios,
                         keep_original=keep_original,
                         original_key=original_key, **config)

    @classmethod
    def _default_config(cls):
        from dxpy.collections.dicts import combine_dicts
        cfg = {
            'keep_original': True,
            'original_key': 'ds1x',
            'rigister_output_with_prefix': False,
            'with_shape_info': False
        }
        return combine_dicts(cfg, super()._default_config())

    def _kernel(self, feeds):
        image_orgin = feeds[NodeKeys.INPUT]
        result = dict()
        for n, v in self.param('down_sample_ratios').items():
            result[n] = DownSampler(image_orgin, v, name=self.name / n)()
        if self.param('keep_original'):
            result[self.param('original_key')] = self.tensor(NodeKeys.INPUT)
        return result


class Padder(Model):
    def __init__(self, input_tensor, padding=None, shape=None, offset=None, name='padder', **config):
        """
        Args:
            padding: str|list of str, value|element value is one of the flollowing:
            -   none: no padding on specific axis
            -   period: period padding
            -   shape: list of shape (target shape)
            -   offset: start index of original image in result image
        """
        super().__init__(name,
                         inputs={NodeKeys.INPUT: input_tensor},
                         padding=padding, shape=shape)

    def _kernel(self, feeds):
        x = feeds[NodeKeys.INPUT]



import tensorflow as tf
from ..graph import Graph, NodeKeys
from contextlib import contextmanager


class Session(Graph):
    def __init__(self, name='session', **config):
        super().__init__(name, **config)
        self.__create_session()

    @classmethod
    def _default_config(cls):
        from dxpy.collections.dicts import combine_dicts
        return combine_dicts({
            'allow_growth': True,
            'log_device_placement': False
        }, super()._default_config())

    def __create_session(self):
        config = tf.ConfigProto()
        if self.param('allow_growth'):
            config.gpu_options.allow_growth = True
        if self.param('log_device_placement'):
            config.log_device_placement = True
        self.register_main_node(tf.Session(config=config))

    def run(self, tensors, feed_dict=None):
        with self.nodes[NodeKeys.MAIN].as_default():
            return self.nodes[NodeKeys.MAIN].run(tensors, feed_dict)

    def run_func(self, func):
        with self.nodes[NodeKeys.MAIN].as_default():
            return func()

    @contextmanager
    def as_default(self):
        with self.nodes[NodeKeys.MAIN].as_default():
            yield self.nodes[NodeKeys.MAIN]

    def post_session_created(self):
        self.as_tensor().run(tf.global_variables_initializer())
        # class SessionWithSupervisor(Graph):
        #     def __init__(self, name, **config):
        #         super(__class__, self).__init__(name, **config)

        #     def __construct(self):
        #         sv_para = {'summary_op': None}
        #         sms = self.c['save']['frequency']
        #         load_step = self.c['save']['load_step']
        #         if sms is not None:
        #             sv_para['save_model_secs'] = sms
        #         if load_step is not None:
        #             sv_para['init_fn'] = load_fn
        #         sv_para['saver'] = save_fn
        #         supervisor = tf.train.Supervisor(**sv_para)
        #         tf_config = tf.ConfigProto(
        #             log_device_placement=config.log.is_show_device_placement)
        #         tf_config.gpu_options.allow_growth = True
        #         sess = supervisor.prepare_or_wait_for_session(config=tf_config)
        #         return {'supervisor': supervisor, 'session': sess}

        #     def _load(self):
        #         pass
        # def supervisor(config, load_fn, save_fn):
        #     sv_para = {'summary_op': None}
        #     sms = config.save.frequency
        #     load_step = config.save.load_step
        #     if sms is not None:
        #         sv_para['save_model_secs'] = sms
        #     if load_step is not None:
        #         sv_para['init_fn'] = load_fn
        #     sv_para['saver'] = save_fn
        #     supervisor = tf.train.Supervisor(**sv_para)
        #     tf_config = tf.ConfigProto(
        #         log_device_placement=config.log.is_show_device_placement)
        #     tf_config.gpu_options.allow_growth = True
        #     sess = supervisor.prepare_or_wait_for_session(config=tf_config)
        #     return {'supervisor': supervisor, 'session': sess}

        # def _set_sesssv(self):
        #         from .supervisor import supervisor
        #         result = supervisor(self.params)



import tensorflow as tf
from dxpy.filesystem import Path
from ..graph import Graph

_instance = None


def get_saver():
    global _instance
    if _instance is None:
        _instance = Saver()
    return _instance


class Saver(Graph):
    def __init__(self, name='/saver', **config):
        super(__class__, self).__init__(name, **config)
        self._saver = None
        self.register_task('save', self.save)
        self.register_task('load', self.load)

    @classmethod
    def _default_config(cls):
        return {'model_dir': './save/',
                'ckpt_name': 'save'}

    def _model_path(self):
        return (Path(self.param('model_dir')) / self.param('ckpt_name')).abs

    def save(self, feeds):
        from ..scalar import global_step
        if self._saver is None:
            self._saver = tf.train.Saver()
        sess = tf.get_default_session()
        step = sess.run(global_step())
        print("[SAVE] model to: {}.".format(self._model_path()))
        self._saver.save(sess, self._model_path(), global_step=step)

    def __resolve_path_load(self, feeds):
        from fs.osfs import OSFS
        import re
        from dxpy.filesystem import Path
        path_check_point = (
            Path(self.param('model_dir', feeds)) / 'checkpoint').abs
        pp = re.compile('^.*: "(.*)".*$')
        ps = re.compile('.*' + self.param('ckpt_name', feeds) + '-([0-9]+)-*')
        paths = []
        with OSFS('/') as fs:
            if not fs.exists(path_check_point):
                return fs.getsyspath(path_check_point), False
            with fs.open(path_check_point) as fin:
                for l in fin.readlines():
                    mat_path = pp.match(l)
                    if mat_path is not None:
                        path_load = mat_path[1]
                        mat_step = ps.match(path_load)
                        if mat_step is not None:
                            paths.append([path_load, int(mat_step[1])])
        step = self.param('step', feeds)
        if step == -1:
            step = max(list(zip(*paths))[1])
        for p, s in paths:
            if s == step:
                return p, True
        return step, False

    def load(self, feeds):
        from ..scalar import global_global_step
        import sys
        if self._saver is None:
            self._saver = tf.train.Saver()
        sess = tf.get_default_session()
        from dxpy.debug import dbgmsg
        dbgmsg(feeds)
        path_load, flag = self.__resolve_path_load(feeds)
        if flag is False:
            if isinstance(path_load, int):
                msg = "[ERROR][LOAD] Save for given step {} not found. Skip restore."
                print(msg.format(path_load), file=sys.stderr)
                return
            else:
                msg = "[ERROR][LOAD] Checkpoint file {} not found. Skip restore."
                print(msg.format(path_load), file=sys.stderr)
                return
        print("[LOAD] model from: {}.".format(path_load))
        self._saver.restore(sess, path_load)



import json
from sqlalchemy import Column, ForeignKey, Integer, String, Boolean, DateTime
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker, scoped_session
from sqlalchemy import create_engine

from dxpy.time.utils import now

Base = declarative_base()


class Database:
    engine = None

    @classmethod
    def create_engine(cls):
        from .config import config as c
        cls.engine = create_engine(c.path_sqllite, echo=c['echo'])

    @classmethod
    def get_or_create_engine(cls):
        if cls.engine is None:
            cls.create_engine()
        return cls.engine

    @classmethod
    def session_maker(cls):
        return sessionmaker(bind=cls.get_or_create_engine())

    @classmethod
    def create(cls):
        Base.metadata.create_all(cls.get_or_create_engine())

    @classmethod
    def clear(cls):
        sess = cls.session_maker()()
        records = sess.query(TaskDB).delete()
        sess.commit()

    @classmethod
    def session(cls):
        return cls.session_maker()()


class TaskDB(Base):
    __tablename__ = 'task'
    id = Column(Integer, primary_key=True)
    desc = Column(String)
    data = Column(String)
    time_create = Column(DateTime)
    time_start = Column(DateTime)
    time_end = Column(DateTime)
    state = Column(String)
    is_root = Column(Boolean)
    ttype = Column(String)
    workdir = Column(String)
    worker = Column(String)
    dependency = Column(String)

    def __init__(self, desc, data, state=None, workdir=None, worker=None, ttype=None, dependency=None, time_create=None, time_start=None, time_end=None, is_root=True):
        self.desc = desc
        self.data = data
        if time_create is None:
            time_create = now()
        self.time_create = time_create
        if state is None:
            from .config import config as c
            state = c['default_state']
        self.state = state
        self.worker = worker
        self.workdir = workdir
        self.ttype = ttype
        self.dependency = dependency
        self.is_root = is_root

    def __repr__(self):
        return '<Task {:d}>'.format(self.id)



from setuptools import setup, find_packages
setup(name='dxl-dxpy',
      version='0.6.3',
      description='Duplicate components library python sub-library.',
      url='https://github.com/Hong-Xiang/dxl',
      author='Hong Xiang',
      author_email='hx.hongxiang@gmail.com',
      license='MIT',
      packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests", "dxpy.tests"]),
      install_requires=[
          'rx',
          'flask_restful',
          'flask_sqlalchemy',
          'graphviz',
          'fs',
          'apscheduler',
          'ruamel.yaml'
      ],
      scripts=['bin/dxpy'],
      zip_safe=False)



from ..base import DatasetTFRecords, NodeKeys
import tensorflow as tf
from ..config import config


class PhantomSinograms(DatasetTFRecords):
    SHAPE_SINOGRAM = [320, 640, 1]
    SHAPE_PHANTOM = [256, 256, 1]

    def __init__(self, name='dataset', **config):
        from dxpy.learn.utils.tensor import ensure_shape
        super().__init__(name, **config)
        self.nodes['image'] = ensure_shape(
            self.nodes['image'], batch_size=self.param('batch_size'))

    @classmethod
    def _parse_example(cls, example):
        return tf.parse_single_example(example, features={
            'shape_phantom': tf.FixedLenFeature([3], tf.int64),
            'shape_sinogram': tf.FixedLenFeature([3], tf.int64),
            'phantom_type': tf.FixedLenFeature([], tf.int64),
            'phantom': tf.FixedLenFeature([], tf.string),
            'sinogram': tf.FixedLenFeature([], tf.string),
        })

    @classmethod
    def _default_config(cls):
        from dxpy.collections.dicts import combine_dicts
        fmt = '/home/hongxwing/Datas/phantom/phantom.{}.tfrecord'
        files = [fmt.format(i) for i in range(1)]
        return combine_dicts({
            'files': files,
            'fields': 'sinogram',
            'batch_size': 32
        }, super()._default_config())

    def _decode_image(self, data):
        if self.param('fields') == 'sinogram':
            return {
                'image': tf.decode_raw(data['sinogram'], tf.uint16),
                'shape': data['shape_sinogram']
            }
        elif self.param('fields') == 'phantom':
            return {
                'image': tf.decode_raw(data['phantom'], tf.uint16),
                'shape': data['shape_phantom']
            }
        else:
            raise ValueError(
                "Unsupported fields: {}.".format(self.param('fields')))

    def _reshape_tensors(self, data):
        if self.param('fields') == 'sinogram':
            shape = self.SHAPE_SINOGRAM
        else:
            shape = self.SHAPE_PHANTOM
        return {'image': tf.reshape(tf.cast(data['image'], tf.float32), shape),
                'shape': data['shape']}

    def _processing(self):
        return (super()._processing()
                .map(self._parse_example)
                .map(self._decode_image)
                .map(self._reshape_tensors)
                .batch(self.c['batch_size'])
                .cache()
                .repeat())



import tensorflow as tf
from ..base import DatasetTFRecords, NodeKeys


class MNISTTFRecords(DatasetTFRecords):
    def __init__(self, name='dataset', **config):
        super(__class__, self).__init__(name, **config)

    # def _pre_processing(self):
    #     self._add_normalizer()

    # def _add_normalizer(self):
    #     from ...model.normalizer import get_normalizer
    #     from ...model.tensor import PlaceHolder
    #     self._normalizer = None
    #     normalizer_class = get_normalizer(self.c['normalization']['method'])
    #     if normalizer_class is not None:
    #         self._normalizer = normalizer_class(self.name / 'normalization',
    #                                             PlaceHolder([28, 28, 1]), lazy_create=True)

    @classmethod
    def _default_config(cls):
        return {
            'files': ['/home/hongxwing/Datas/mnist/mnist.train.tfrecord'],
            'batch_size': 32,
            'one_hot': True,
        }

    @classmethod
    def _parse_example(cls, example):
        return tf.parse_single_example(example, features={
            'image': tf.FixedLenFeature([], tf.string),
            'shape': tf.FixedLenFeature([2], tf.int64),
            'label': tf.FixedLenFeature([1], tf.int64)
        })

    @classmethod
    def _decode_image(cls, data):
        return {'image': tf.decode_raw(data['image'], tf.uint8),
                'shape': data['shape'],
                'label': data['label']}

    @classmethod
    def _change_dtype(cls, data):
        return {'image': tf.to_float(data['image']),
                'shape': data['shape'],
                'label': data['label']}

    @classmethod
    def _reshape_tensors(cls, data):
        return {'image': tf.reshape(data['image'], [28, 28, 1]),
                'shape': tf.concat([data['shape'], [1]], axis=0),
                'label': tf.reshape(data['label'], [])}

    def _normalize(self, data):
        if self._normalizer is None:
            image = data['image']
        else:
            image = self._normalizer(data['image'])[NodeKeys.MAIN]
        return {'image': image,
                'shape': data['shape'],
                'label': data['label']}
        return data

    def _maybe_onehot(self, data):
        result = {
            'image': data['image'],
            'shape': data['shape']
        }
        if self.c['one_hot']:
            result.update({'label': tf.one_hot(data['label'], depth=10)})
        else:
            result.update({'label': data['label']})
        return result

    def _processing(self):
        dataset = super()._processing()
        return (dataset
                .map(self._parse_example)
                .map(self._decode_image)
                .map(self._change_dtype)
                .map(self._reshape_tensors)
                # .map(self._normalize)
                .map(self._maybe_onehot)
                .batch(self.c['batch_size'])
                .cache()
                .repeat())


from ..base import Graph


class MNISTLoadAll(Graph):
    def __init__(self, name='/dataset', **config):
        # from ..preprocessing.normalizer import SelfMinMax

        super(__class__, self).__init__(name, **config)
        self._normalizer = None
        method = self.c['normalization']['method'].lower()
        if method != 'pass':
            self._normalizer = get_normalizer(
                method)(self.name / 'normalization', )
        self.register_main_node(self.__load_data())
        self.register_node('image', self.as_tensor()['image'])
        self.register_node('label', self.as_tensor()['label'])

    def get_feed_dict(self):
        return tf.get_default_session().run(
            {'image': self.tensor('image'), 'label': self.tensor('label')})

    @classmethod
    def _default_config(cls):
        return {
            'batch_size': 32,
        }

    def post_session_created(self):
        tf.get_default_session().run(self._iter_init,
                                     feed_dict={
                                         self.tensor('images'): self.images,
                                         self.tensor('labels'): self.labels
                                     })

    def _normalize(self, image):
        if self._normalizer is None:
            return image
        else:
            return self._normalizer(image)['data']

    def __load_data(self):
        from tensorflow.examples.tutorials.mnist import input_data
        from tensorflow.contrib.data import Dataset
        mnist = input_data.read_data_sets(
            '/home/hongxwing/Datas/mnist/tfdefault', one_hot=True)
        self.images = mnist.train.images
        self.labels = mnist.train.labels

        # def gen_image():
        #     for i in range(images.shape[0]):
        #         yield {'image': images[i, ...], 'label': label[i, ...]}
        with tf.name_scope(self.basename):
            images_tensor = tf.placeholder(tf.float32, self.images.shape,
                                           'mnist_images')
            self.register_node('images', images_tensor)
            labels_tensor = tf.placeholder(tf.int64, self.labels.shape,
                                           'mnist_label')
            self.register_node('labels', labels_tensor)
            dataset_image = (Dataset.from_tensor_slices(images_tensor)
                             .map(lambda img: tf.reshape(img, [28, 28, 1]))
                             .map(self._normalize))
            dataset_label = Dataset.from_tensor_slices(labels_tensor)
            dataset = Dataset.zip(
                {'image': dataset_image, 'label': dataset_label})
            dataset = dataset.batch(self.param('batch_size')).repeat()
            iterator = dataset.make_initializable_iterator()
            self._iter_init = iterator.initializer
            next_element = iterator.get_next()
        return next_element



from dxpy.configs import ModuleConfigs
from dxpy.filesystem import Path
import os
# config = ModuleConfigs('learn', 'dxpy', os.getcwd()).get_config()
config = dict()


def clear_config():
    global config
    config = dict()



import json

import rx

from dxpy.time.utils import strf, strp

from ..exceptions import InvalidJSONForTask, TaskDatabaseConnectionError, TaskNotFoundError
from ..model import Database, TaskDB


def check_json(s, is_with_id=False):
    dct = json.loads(s)
    if not dct.get('__task__'):
        raise InvalidJSONForTask(
            "'__task__' not found or is False for JSON: {}".format(s))

    def check_key(dct, key, value_type=None):
        if not key in dct:
            raise InvalidJSONForTask(
                "Required key: {key} is not found in JSON string: {s}".format(key=key, s=s))
        if value_type is not None:
            if not isinstance(dct[key], value_type):
                raise InvalidJSONForTask(
                    "Wrong type for key: {k} with value: {v}".format(k=key, v=dct[k]))

    if is_with_id:
        check_key(dct, 'id', int)
    check_key(dct, 'desc', str)
    check_key(dct, 'data', dict)
    check_key(dct, 'time_stamp', dict)
    check_key(dct['time_stamp'], 'create', (str, type(None)))
    check_key(dct['time_stamp'], 'start', (str, type(None)))
    check_key(dct['time_stamp'], 'end', (str, type(None)))
    check_key(dct, 'state', str)
    check_key(dct, 'is_root', bool)
    check_key(dct, 'worker', str)
    check_key(dct, 'type', str)
    check_key(dct, 'workdir', str)
    check_key(dct, 'dependency', list)


def db2json(task):
    return json.dumps({
        '__task__': True,
        'id': task.id,
        'desc': task.desc,
        'data': json.loads(task.data),
        'worker': task.worker,
        'type': task.ttype,
        'workdir': task.workdir,
        'dependency': json.loads(task.dependency),
        'time_stamp': {
            'create': strf(task.time_create),
            'start': strf(task.time_start),
            'end': strf(task.time_end)},
        'state': task.state,
        'is_root': task.is_root
    })


def json2db_new(s):
    check_json(s)
    dct = json.loads(s)

    return TaskDB(desc=dct['desc'],
                  data=json.dumps(dct['data']),
                  state=dct['state'],
                  workdir=dct['workdir'],
                  worker=dct['worker'],
                  ttype=dct['type'],
                  dependency=json.dumps(dct['dependency']),
                  time_create=strp(dct['time_stamp']['create']),
                  is_root=dct['is_root'])


def jsonit2db(s):
    dct = json.loads(s)


def jsondb2it(s):
    pass


class Service:
    session = None

    @classmethod
    def create_session(cls):
        cls.session = Database.session()

    @classmethod
    def get_or_create_session(cls, path=None):
        if cls.session is None:
            cls.create_session()
        return cls.session

    @classmethod
    def clear_session(cls):
        if cls.session is not None:
            cls.session.close()
        cls.session = None

    @classmethod
    def create(cls, task_json: str) -> int:
        """
        Create a task record in database.
        """
        task_db = json2db_new(task_json)
        cls.get_or_create_session().add(task_db)
        cls.get_or_create_session().commit()
        task_db = cls.get_or_create_session().query(TaskDB).get(task_db.id)
        return task_db.id

    @classmethod
    def read_taskdb(cls, tid):
        task_db = cls.get_or_create_session().query(TaskDB).get(tid)
        if task_db is None:
            raise TaskNotFoundError(tid)
        else:
            return task_db

    @classmethod
    def read(cls, tid=None):
        """
        Raises:
        - `TaskNotFoundError`
        """
        return db2json(cls.read_taskdb(tid))

    @classmethod
    def read_all(cls, filter_func=None) -> 'rx.Observable<json str>':
        """
        Returns JSON serilized query results;

        Returns:
        - `str`: JSON loadable observalbe

        Raises:
        - None
        """
        if filter_func is None:
            def filter_func(x): return True
        return (rx.Observable
                .from_(cls.get_or_create_session().query(TaskDB).all())
                .filter(filter_func)
                .map(db2json))

    @classmethod
    def json2db_update(cls, s):
        check_json(s, is_with_id=True)
        dct = json.loads(s)
        taskdb = cls.read_taskdb(dct['id'])
        taskdb.desc = dct['desc']
        taskdb.data = json.dumps(dct['data'])
        taskdb.state = dct['state']
        taskdb.worker = dct['worker']
        taskdb.workdir = dct['workdir']
        taskdb.ttype = dct['type']
        taskdb.dependency = json.dumps(dct['dependency'])
        taskdb.time_create = strp(dct['time_stamp']['create'])
        taskdb.time_start = strp(dct['time_stamp']['start'])
        taskdb.time_end = strp(dct['time_stamp']['end'])
        taskdb.is_root = dct['is_root']
        return taskdb

    @classmethod
    def update(cls, task_json):
        cls.json2db_update(task_json)
        cls.get_or_create_session().commit()

    @classmethod
    def delete(cls, tid):
        t = cls.read_taskdb(tid)
        if t is not None:
            cls.get_or_create_session().delete(t)
            cls.get_or_create_session().commit()



import numpy as np
import tensorflow as tf


def ensure_shape(input_tensor, shape=None, *, batch_size=None):
    """
    Reshpe input_tensor to given shape.
    """
    if isinstance(input_tensor, np.ndarray):
        return np.copy(input_tensor)
    with tf.name_scope('shape_ensure'):
        shape_input = input_tensor.shape.as_list()
        if shape is None:
            shape = shape_input
        if shape[0] is None:
            shape[0] = batch_size
        return tf.reshape(input_tensor, shape)
